import torch
import torch.nn as nn


class SLAF(nn.Module):
    def __init__(self, k=2):
        super().__init__()
        self.k = k
        self.coeff = nn.ParameterList(
            [nn.Parameter(torch.tensor(1.0)) for i in range(k)])

    def forward(self, x):
        out = sum([self.coeff[k] * torch.pow(x, k) for k in range(self.k)])
        return out



from models import BaseModel
import torch
import torch.nn as nn
import torch.nn.functional as F


class LinearMNIST(BaseModel):
    def __init__(self, activation: nn.Module, in_ch: int=1):
        super().__init__(activation)

        self.linear1 = nn.Sequential(
            nn.Linear(in_ch * 28 * 28, 512),
            activation(),
        )

        self.linear2 = nn.Sequential(
            nn.Linear(512, 10),
            nn.LogSoftmax(dim=-1)
        )

    def forward(self, x):
        x = x.view(-1, 28 * 28)

        x = self.linear1(x)

        x = self.linear2(x)

        return x



import torch
from torch import nn


class Swish(torch.autograd.Function):

    @staticmethod
    def forward(ctx, i):
        result = i * i.sigmoid()
        ctx.save_for_backward(result, i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        result, i = ctx.saved_variables
        sigmoid_x = i.sigmoid()
        return grad_output * (result + sigmoid_x * (1 - result))


swish = Swish.apply


class Swish_module(nn.Module):
    def forward(self, x):
        return swish(x)


ACTIVATION_FUNCTIONS = dict({
    # pau
    "pade_optimized_leakyrelu_abs": "pade_optimized_leakyrelu_abs",
    "sigmoid": nn.Sigmoid(),
    "relu": nn.ReLU, "selu": nn.SELU, "leakyrelu": nn.LeakyReLU, "celu": nn.CELU,
    "elu": nn.ELU,
    "tanh": nn.Tanh,
    "relu6": nn.ReLU6,
    "swish": Swish_module,
    "softplus": nn.Softplus,
    "prelu": nn.PReLU,
    "rrelu": nn.RReLU})

from activations.pau.cuda.python_imp.Pade import PADEACTIVATION_Function_based, PADEACTIVATION_F_cpp, \
    PADEACTIVATION_F_abs_cpp


class activationfunc():
    def __init__(self, selected_activation_func):
        self.selected_activation_func = selected_activation_func

        assert "pade" in selected_activation_func or selected_activation_func in ACTIVATION_FUNCTIONS, "unknown activation function %s" % selected_activation_func

    def get_activationfunc(self):
        if "pade" in self.selected_activation_func:
            PADEACTIVATION_F_abs_cpp.config_cuda(5, 4, 0.)
            init_coefficients = self.selected_activation_func.replace("_abs", "").replace("_cuda", "")
            if "_abs" in self.selected_activation_func:
                return PADEACTIVATION_Function_based(init_coefficients=init_coefficients,
                                                     act_func_cls=PADEACTIVATION_F_abs_cpp)
            else:
                return PADEACTIVATION_Function_based(init_coefficients=init_coefficients,
                                                     act_func_cls=PADEACTIVATION_F_cpp)
        else:
            return ACTIVATION_FUNCTIONS[self.selected_activation_func]()


def PAU():
    PADEACTIVATION_F_abs_cpp.config_cuda(5, 4, 0.)
    return PADEACTIVATION_Function_based(init_coefficients="pade_optimized_leakyrelu",
                                         act_func_cls=PADEACTIVATION_F_abs_cpp)



import numpy as np
from visdom import Visdom
PORT = 8097

_WINDOW_CASH = {}
_ENV_CASH = {}


def _vis(env="main"):
    if env not in _ENV_CASH:
        _ENV_CASH[env] = Visdom(env=env, port=PORT)
    return _ENV_CASH[env]


def visualize_losses(loss: dict, title: str, env="main", epoch=0):
    legend = list()
    scalars = list()
    dash = []
    flag = 0
    for k, v in loss.items():
        legend.append(k)
        scalars.append(v)
        if flag % 3 == 0:
            dash.append("solid")
        elif flag % 3 == 1:
            dash.append("dash")
        elif flag % 3 == 2:
            dash.append("dashdot")
        flag += 1
    dash = np.asarray(dash)
    options = dict(
        width=1200,
        height=600,
        xlabel="Epochs",
        title=title,
        marginleft=30,
        marginright=30,
        marginbottom=80,
        margintop=30,
        legend=legend,
        fillarea=False,
        dash=dash,
    )
    if title in _WINDOW_CASH:
        _vis(env).line(Y=[scalars], X=[epoch],
                       win=_WINDOW_CASH[title], update="append", opts=options)
    else:
        _WINDOW_CASH[title] = _vis(env).line(
            Y=[scalars], X=[epoch], opts=options)


def visualize_accuracy(loss: dict, title: str, env="main", epoch=0):
    legend = list()
    scalars = list()
    dash = []
    flag = 0
    for k, v in loss.items():
        legend.append(k)
        scalars.append(v)
        if flag % 3 == 0:
            dash.append("solid")
        elif flag % 3 == 1:
            dash.append("dash")
        elif flag % 3 == 2:
            dash.append("dashdot")
        flag += 1
    dash = np.asarray(dash)
    options = dict(
        width=1200,
        height=600,
        xlabel="Epochs",
        ylabel="%",
        title=title,
        marginleft=30,
        marginright=30,
        marginbottom=80,
        margintop=30,
        legend=legend,
        fillarea=False,
        dash=dash,
    )
    if title in _WINDOW_CASH:
        _vis(env).line(Y=[scalars], X=[epoch],
                       win=_WINDOW_CASH[title], update="append", opts=options)
    else:
        _WINDOW_CASH[title] = _vis(env).line(
            Y=[scalars], X=[epoch], opts=options)



import random
from collections import namedtuple

import numpy as np
import plotly
import plotly.graph_objs as go
import plotly.plotly as py

with open('visualize/color_map.txt', "r") as f:
    color_map = [line.split(" ")[0] for line in f]

class ContinuousErrorBars(object):
    def __init__(self, dicts: dict):
        """ dicts:
                -> ReLU:
                    -> runtime 1
                    -> runtime 2
                    -> ...
                -> ...
        """
        self.dicts = dicts

    def draw(self, filename, ticksuffix=""):
        # X
        for k, v in self.dicts.items():
            x_len = len(v[0])
            break

        x = list(range(1, x_len + 1))
        x_rev = x[::-1]

        # LOOP Y
        y = {k: [] for k in self.dicts.keys()}
        y_lower = {k: [] for k in self.dicts.keys()}
        y_upper = {k: [] for k in self.dicts.keys()}

        for k, v in self.dicts.items():
            for i in range(x_len):
                d = []
                for t in range(len(v)):
                    d.append(v[t][i])
                y[k].append(np.mean(d))
                y_lower[k].append(np.min(d))
                y_upper[k].append(np.max(d))

        y_lower = {k: v[::-1] for k, v in y_lower.items()}

        # TRACE
        data = []
        # UPPER AND LOWER
        for i, k in enumerate(self.dicts.keys()):
            trace = go.Scatter(
                x=x+x_rev,
                y=y_upper[k] + y_lower[k],
                fill="tozerox",
                fillcolor="rbga({}, 0.2)".format(color_map[i]),
                line=dict(color="rgba(255,255,255,0)"),
                mode="none",
                showlegend=False,
                name=k
            )
            data.append(trace)

            # MEAN
            trace = go.Scatter(
                x=x,
                y=y[k],
                line=dict(color="rgb({})".format(color_map[i])),
                mode="lines",
                name=k
            )
            data.append(trace)

        layout = go.Layout(
            paper_bgcolor="rgb(255, 255, 255)",
            plot_bgcolor="rgb(255, 255, 255)",
            xaxis=dict(
                gridcolor="rgb(229, 229, 229)",
                range=[1, x_len],
                showgrid=True,
                showline=False,
                showticklabels=True,
                tickcolor="rgb(127, 127, 127)",
                ticks="outside",
                zeroline=False,
                ticktext="Epochs",
            ),
            yaxis=dict(
                gridcolor="rgb(229, 229, 229)",
                showgrid=True,
                showline=False,
                showticklabels=True,
                tickcolor="rgb(127, 127, 127)",
                ticks="outside",
                zeroline=False,
                ticksuffix=ticksuffix
            ),
        )

        fig = go.Figure(data=data, layout=layout)
        plotly.offline.plot(fig, filename=filename)



import torch
import torch.nn as nn


class GELU(nn.Module):
    """https://mlfromscratch.com/activation-functions-explained/#/
    """

    def __init__(self):
        super().__init__()

    def forward(self, x):
        return 0.5 * x * (1 + torch.tanh(0.7979946 * (x + 0.044715 * torch.pow(x, 3))))



import torch
import torch.nn as nn


class Mixture(nn.Module):
    def __init__(self,):
        super().__init__()
        self.p = nn.Parameter(torch.tensor([0.0]))

    def forward(self, x):
        return self.p * x + (1-self.p) * torch.relu(x)



import torch
import torch.nn as nn
import torch.nn.functional as F


class AReLU(nn.Module):
    def __init__(self, alpha=0.90, beta=2.0):
        super().__init__()
        self.alpha = nn.Parameter(torch.tensor([alpha]))
        self.beta = nn.Parameter(torch.tensor([beta]))

    def forward(self, input):
        alpha = torch.clamp(self.alpha, min=0.01, max=0.99)
        beta = 1 + torch.sigmoid(self.beta)

        return F.relu(input) * beta - F.relu(-input) * alpha



import torch
import torch.nn as nn


class APL(nn.Module):
    def __init__(self, s=1):
        super().__init__()

        self.a = nn.ParameterList(
            [nn.Parameter(torch.tensor(0.2)) for _ in range(s)])
        self.b = nn.ParameterList(
            [nn.Parameter(torch.tensor(0.5)) for _ in range(s)])
        self.s = s

    def forward(self, x):
        part_1 = torch.clamp_min(x, min=0.0)
        part_2 = 0
        for i in range(self.s):
            part_2 += self.a[i] * torch.clamp_min(-x+self.b[i], min=0)

        return part_1 + part_2



import torch
import torch.nn as nn
import torch.nn.functional as F
from models import BaseModel


def conv_block(in_plane, out_plane, kernel_size, activation):
    return nn.Sequential(
        nn.Conv2d(in_channels=in_plane, out_channels=out_plane,
                  kernel_size=kernel_size),
        nn.MaxPool2d(2),
        activation(),
    )


class ConvMNIST(BaseModel):
    def __init__(self, activation: nn.Module, in_ch: int=1):
        super().__init__(activation)

        self.conv_block1 = conv_block(
            in_ch, 10, kernel_size=5, activation=activation)
        self.conv_block2 = conv_block(
            10, 20, kernel_size=5, activation=activation)
        self.conv_block3 = conv_block(
            20, 40, kernel_size=3, activation=activation)

        self.fc = nn.Sequential(
            nn.Linear(40, 10),
            nn.LogSoftmax(dim=-1),
        )

    def forward(self, x):
        x = self.conv_block1(x)
        x = self.conv_block2(x)
        x = self.conv_block3(x)

        x = x.view(-1, 40)

        x = self.fc(x)

        return x



#!/usr/bin/env python3

"""
**Description**

A set of commonly used models for meta-learning vision tasks.
For simplicity, all models' `forward` conform to the following API:

~~~python
def forward(self, x):
    x = self.features(x)
    x = self.classifier(x)
    return x
~~~
"""

import torch
import learn2learn as l2l

from scipy.stats import truncnorm


def truncated_normal_(tensor, mean=0.0, std=1.0):
    # PT doesn't have truncated normal.
    # https://discuss.pytorch.org/t/implementing-truncated-normal-initializer/4778/18
    values = truncnorm.rvs(-2, 2, size=tensor.shape)
    values = mean + std * values
    tensor.copy_(torch.from_numpy(values))
    return tensor


def fc_init_(module):
    if hasattr(module, 'weight') and module.weight is not None:
        truncated_normal_(module.weight.data, mean=0.0, std=0.01)
    if hasattr(module, 'bias') and module.bias is not None:
        torch.nn.init.constant_(module.bias.data, 0.0)
    return module


def maml_init_(module):
    torch.nn.init.xavier_uniform_(module.weight.data, gain=1.0)
    torch.nn.init.constant_(module.bias.data, 0.0)
    return module


class LinearBlock(torch.nn.Module):

    def __init__(self, input_size, output_size):
        super(LinearBlock, self).__init__()
        self.relu = torch.nn.ReLU()
        self.normalize = torch.nn.BatchNorm1d(
            output_size,
            affine=True,
            momentum=0.999,
            eps=1e-3,
            track_running_stats=False,
        )
        self.linear = torch.nn.Linear(input_size, output_size)
        fc_init_(self.linear)

    def forward(self, x):
        x = self.linear(x)
        x = self.normalize(x)
        x = self.relu(x)
        return x


class ConvBlock(torch.nn.Module):

    def __init__(self,
                 afs,
                 in_channels,
                 out_channels,
                 kernel_size,
                 max_pool=True,
                 max_pool_factor=1.0):
        super(ConvBlock, self).__init__()
        stride = (int(2 * max_pool_factor), int(2 * max_pool_factor))
        if max_pool:
            self.max_pool = torch.nn.MaxPool2d(
                kernel_size=stride,
                stride=stride,
                ceil_mode=False,
            )
            stride = (1, 1)
        else:
            self.max_pool = lambda x: x
        self.normalize = torch.nn.BatchNorm2d(
            out_channels,
            affine=True,
            # eps=1e-3,
            # momentum=0.999,
            # track_running_stats=False,
        )
        torch.nn.init.uniform_(self.normalize.weight)
        # self.relu = torch.nn.ReLU()
        self.afs = afs()

        self.conv = torch.nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=1,
            bias=True,
        )
        maml_init_(self.conv)

    def forward(self, x):
        x = self.conv(x)
        x = self.normalize(x)
        # x = self.relu(x)
        x = self.afs(x)
        x = self.max_pool(x)
        return x


class ConvBase(torch.nn.Sequential):

    # NOTE:
    #     Omniglot: hidden=64, channels=1, no max_pool
    #     MiniImagenet: hidden=32, channels=3, max_pool

    def __init__(self,
                 afs,
                 output_size,
                 hidden=64,
                 channels=1,
                 max_pool=False,
                 layers=4,
                 max_pool_factor=1.0):
        core = [ConvBlock(
            afs,
            channels,
            hidden,
            (3, 3),
            max_pool=max_pool,
            max_pool_factor=max_pool_factor),
        ]
        for _ in range(layers - 1):
            core.append(ConvBlock(hidden,
                                  hidden,
                                  kernel_size=(3, 3),
                                  max_pool=max_pool,
                                  max_pool_factor=max_pool_factor))
        super(ConvBase, self).__init__(*core)


class OmniglotFC(torch.nn.Module):
    """

    [[Source]](https://github.com/learnables/learn2learn/blob/master/learn2learn/vision/models.py)

    **Description**

    The fully-connected network used for Omniglot experiments, as described in Santoro et al, 2016.

    **References**

    1. Santoro et al. 2016. “Meta-Learning with Memory-Augmented Neural Networks.” ICML.

    **Arguments**

    * **input_size** (int) - The dimensionality of the input.
    * **output_size** (int) - The dimensionality of the output.
    * **sizes** (list, *optional*, default=None) - A list of hidden layer sizes.

    **Example**
    ~~~python
    net = OmniglotFC(input_size=28**2,
                     output_size=10,
                     sizes=[64, 64, 64])
    ~~~

    """

    def __init__(self, input_size, output_size, sizes=None):
        super(OmniglotFC, self).__init__()
        if sizes is None:
            sizes = [256, 128, 64, 64]
        layers = [LinearBlock(input_size, sizes[0]), ]
        for s_i, s_o in zip(sizes[:-1], sizes[1:]):
            layers.append(LinearBlock(s_i, s_o))
        layers = torch.nn.Sequential(*layers)
        self.features = torch.nn.Sequential(
            l2l.nn.Flatten(),
            layers,
        )
        self.classifier = fc_init_(torch.nn.Linear(sizes[-1], output_size))
        self.input_size = input_size

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)
        return x


class OmniglotCNN(torch.nn.Module):
    """

    [Source](https://github.com/learnables/learn2learn/blob/master/learn2learn/vision/models.py)

    **Description**

    The convolutional network commonly used for Omniglot, as described by Finn et al, 2017.

    This network assumes inputs of shapes (1, 28, 28).

    **References**

    1. Finn et al. 2017. “Model-Agnostic Meta-Learning for Fast Adaptation of Deep Networks.” ICML.

    **Arguments**

    * **output_size** (int) - The dimensionality of the network's output.
    * **hidden_size** (int, *optional*, default=64) - The dimensionality of the hidden representation.
    * **layers** (int, *optional*, default=4) - The number of convolutional layers.

    **Example**
    ~~~python
    model = OmniglotCNN(output_size=20, hidden_size=128, layers=3)
    ~~~

    """

    def __init__(self, output_size=5, hidden_size=64, layers=4):
        super(OmniglotCNN, self).__init__()
        self.hidden_size = hidden_size
        self.base = ConvBase(output_size=hidden_size,
                             hidden=hidden_size,
                             channels=1,
                             max_pool=False,
                             layers=layers)
        self.features = torch.nn.Sequential(
            l2l.nn.Lambda(lambda x: x.view(-1, 1, 28, 28)),
            self.base,
            l2l.nn.Lambda(lambda x: x.mean(dim=[2, 3])),
            l2l.nn.Flatten(),
        )
        self.classifier = torch.nn.Linear(hidden_size, output_size, bias=True)
        self.classifier.weight.data.normal_()
        self.classifier.bias.data.mul_(0.0)

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)
        return x


class MiniImagenetCNN(torch.nn.Module):
    """

    [[Source]](https://github.com/learnables/learn2learn/blob/master/learn2learn/vision/models.py)

    **Description**

    The convolutional network commonly used for MiniImagenet, as described by Ravi et Larochelle, 2017.

    This network assumes inputs of shapes (3, 84, 84).

    **References**

    1. Ravi and Larochelle. 2017. “Optimization as a Model for Few-Shot Learning.” ICLR.

    **Arguments**

    * **output_size** (int) - The dimensionality of the network's output.
    * **hidden_size** (int, *optional*, default=32) - The dimensionality of the hidden representation.
    * **layers** (int, *optional*, default=4) - The number of convolutional layers.

    **Example**
    ~~~python
    model = MiniImagenetCNN(output_size=20, hidden_size=128, layers=3)
    ~~~
    """

    def __init__(self, output_size, afs, hidden_size=32, layers=4):
        super(MiniImagenetCNN, self).__init__()
        base = ConvBase(
            afs,
            output_size=hidden_size,
            hidden=hidden_size,
            channels=3,
            max_pool=True,
            layers=layers,
            max_pool_factor=4 // layers,
        )
        self.features = torch.nn.Sequential(
            base,
            l2l.nn.Flatten(),
        )
        self.classifier = torch.nn.Linear(
            25 * hidden_size,
            output_size,
            bias=True,
        )
        maml_init_(self.classifier)
        self.hidden_size = hidden_size

    def forward(self, x):
        x = self.features(x)
        x = self.classifier(x)
        return x



from .apl import APL
from .arelu import AReLU
from .gelu import GELU
from .maxout import Maxout
from .mixture import Mixture
from .slaf import SLAF
from .swish import Swish
from torch.nn import ReLU, ReLU6, Sigmoid, LeakyReLU, ELU, PReLU, SELU, Tanh, RReLU, CELU, Softplus


__class_dict__ = {key: var for key, var in locals().items()
                  if isinstance(var, type)}
try:
    from .pau.utils import PAU
    __class_dict__["PAU"] = PAU
except Exception:
    raise NotImplementedError("")


__version__ = "0.1.0"


import torch
import torch.nn as nn


class Swish_fun(torch.autograd.Function):

    @staticmethod
    def forward(ctx, i):
        result = i * i.sigmoid()
        ctx.save_for_backward(result, i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        result, i = ctx.saved_variables
        sigmoid_x = i.sigmoid()
        return grad_output * (result + sigmoid_x * (1 - result))


swish = Swish_fun.apply


class Swish(nn.Module):
    def forward(self, x):
        return swish(x)



import torch
import torch.nn as nn


class Maxout(nn.Module):
    """https://github.com/pytorch/pytorch/issues/805#issuecomment-460385007
    """

    def __init__(self, pool_size=1):
        super().__init__()
        self._pool_size = pool_size

    def forward(self, x):
        assert x.shape[1] % self._pool_size == 0, \
            'Wrong input last dim size ({}) for Maxout({})'.format(
                x.shape[1], self._pool_size)
        m, i = x.view(*x.shape[:1], x.shape[1] // self._pool_size,
                      self._pool_size, *x.shape[2:]).max(2)
        return m



'''
python setup.py sdist bdist_wheel
python -m twine upload dist/*
'''
import os

from setuptools import find_packages, setup

from activations import __version__

install_requires = "torch"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    install_requires=install_requires,
    name="activations",
    version=__version__,
    author="densechen",
    author_email="densechen@foxmail.com",
    description="activations: a package contains different kinds of activation functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/densechen/AReLU",
    download_url = 'https://github.com/densechen/AReLU/archive/master.zip',
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    python_requires='>=3.6',
)



import torch
import torch.nn as nn


class BaseModel(nn.Module):
    """ Base Model for implementation different models.
    """

    def __init__(self, activation: nn.Module, in_ch: int=1):
        super().__init__()
